"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class AVTransport1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'AVTransport1' service.
    """

    SERVICE_MANUFACTURER = 'SonosInc'
    SERVICE_TYPE = 'urn:schemas-upnp-org:service:AVTransport:1'

    SERVICE_DEFAULT_VARIABLES = {
        "AVTransportURI": { "data_type": "string", "default": None, "allowed_list": None},
        "AVTransportURIMetaData": { "data_type": "string", "default": None, "allowed_list": None},
        "AbsoluteCounterPosition": { "data_type": "i4", "default": None, "allowed_list": None},
        "AbsoluteTimePosition": { "data_type": "string", "default": None, "allowed_list": None},
        "AlarmIDRunning": { "data_type": "ui4", "default": None, "allowed_list": None},
        "AlarmLoggedStartTime": { "data_type": "string", "default": None, "allowed_list": None},
        "AlarmRunning": { "data_type": "boolean", "default": None, "allowed_list": None},
        "CurrentCrossfadeMode": { "data_type": "boolean", "default": None, "allowed_list": None},
        "CurrentMediaDuration": { "data_type": "string", "default": None, "allowed_list": None},
        "CurrentPlayMode": { "data_type": "string", "default": "NORMAL", "allowed_list": "['NORMAL', 'REPEAT_ALL', 'REPEAT_ONE', 'SHUFFLE_NOREPEAT', 'SHUFFLE', 'SHUFFLE_REPEAT_ONE']"},
        "CurrentRecordQualityMode": { "data_type": "string", "default": None, "allowed_list": None},
        "CurrentSection": { "data_type": "ui4", "default": None, "allowed_list": None},
        "CurrentTrack": { "data_type": "ui4", "default": None, "allowed_list": None},
        "CurrentTrackDuration": { "data_type": "string", "default": None, "allowed_list": None},
        "CurrentTrackMetaData": { "data_type": "string", "default": None, "allowed_list": None},
        "CurrentTrackURI": { "data_type": "string", "default": None, "allowed_list": None},
        "CurrentTransportActions": { "data_type": "string", "default": None, "allowed_list": None},
        "CurrentValidPlayModes": { "data_type": "string", "default": None, "allowed_list": None},
        "DirectControlAccountID": { "data_type": "string", "default": None, "allowed_list": None},
        "DirectControlClientID": { "data_type": "string", "default": None, "allowed_list": None},
        "DirectControlIsSuspended": { "data_type": "boolean", "default": None, "allowed_list": None},
        "EnqueuedTransportURI": { "data_type": "string", "default": None, "allowed_list": None},
        "EnqueuedTransportURIMetaData": { "data_type": "string", "default": None, "allowed_list": None},
        "MuseSessions": { "data_type": "string", "default": None, "allowed_list": None},
        "NextAVTransportURI": { "data_type": "string", "default": None, "allowed_list": None},
        "NextAVTransportURIMetaData": { "data_type": "string", "default": None, "allowed_list": None},
        "NextTrackMetaData": { "data_type": "string", "default": None, "allowed_list": None},
        "NextTrackURI": { "data_type": "string", "default": None, "allowed_list": None},
        "NumberOfTracks": { "data_type": "ui4", "default": None, "allowed_list": None},
        "PlaybackStorageMedium": { "data_type": "string", "default": None, "allowed_list": "['NONE', 'NETWORK']"},
        "PossiblePlaybackStorageMedia": { "data_type": "string", "default": None, "allowed_list": None},
        "PossibleRecordQualityModes": { "data_type": "string", "default": None, "allowed_list": None},
        "PossibleRecordStorageMedia": { "data_type": "string", "default": None, "allowed_list": None},
        "QueueUpdateID": { "data_type": "ui4", "default": None, "allowed_list": None},
        "RecordMediumWriteStatus": { "data_type": "string", "default": None, "allowed_list": None},
        "RecordStorageMedium": { "data_type": "string", "default": None, "allowed_list": "['NONE']"},
        "RelativeCounterPosition": { "data_type": "i4", "default": None, "allowed_list": None},
        "RelativeTimePosition": { "data_type": "string", "default": None, "allowed_list": None},
        "RestartPending": { "data_type": "boolean", "default": None, "allowed_list": None},
        "SleepTimerGeneration": { "data_type": "ui4", "default": None, "allowed_list": None},
        "SnoozeRunning": { "data_type": "boolean", "default": None, "allowed_list": None},
        "TransportErrorDescription": { "data_type": "string", "default": None, "allowed_list": None},
        "TransportErrorHttpCode": { "data_type": "string", "default": None, "allowed_list": None},
        "TransportErrorHttpHeaders": { "data_type": "string", "default": None, "allowed_list": None},
        "TransportErrorURI": { "data_type": "string", "default": None, "allowed_list": None},
        "TransportPlaySpeed": { "data_type": "string", "default": None, "allowed_list": "['1']"},
        "TransportState": { "data_type": "string", "default": None, "allowed_list": "['STOPPED', 'PLAYING', 'PAUSED_PLAYBACK', 'TRANSITIONING']"},
        "TransportStatus": { "data_type": "string", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {
        "LastChange": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_AddMultipleURIsToQueue(self, InstanceID, UpdateID, NumberOfURIs, EnqueuedURIs, EnqueuedURIsMetaData, ContainerURI, ContainerMetaData, DesiredFirstTrackNumberEnqueued, EnqueueAsNext, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AddMultipleURIsToQueue action.

            :returns: "FirstTrackNumberEnqueued", "NumTracksAdded", "NewQueueLength", "NewUpdateID"
        """
        arguments = {
            "InstanceID": InstanceID,
            "UpdateID": UpdateID,
            "NumberOfURIs": NumberOfURIs,
            "EnqueuedURIs": EnqueuedURIs,
            "EnqueuedURIsMetaData": EnqueuedURIsMetaData,
            "ContainerURI": ContainerURI,
            "ContainerMetaData": ContainerMetaData,
            "DesiredFirstTrackNumberEnqueued": DesiredFirstTrackNumberEnqueued,
            "EnqueueAsNext": EnqueueAsNext,
        }

        out_params = self.call_action("AddMultipleURIsToQueue", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("FirstTrackNumberEnqueued", "NumTracksAdded", "NewQueueLength", "NewUpdateID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_AddURIToQueue(self, InstanceID, EnqueuedURI, EnqueuedURIMetaData, DesiredFirstTrackNumberEnqueued, EnqueueAsNext, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AddURIToQueue action.

            :returns: "FirstTrackNumberEnqueued", "NumTracksAdded", "NewQueueLength"
        """
        arguments = {
            "InstanceID": InstanceID,
            "EnqueuedURI": EnqueuedURI,
            "EnqueuedURIMetaData": EnqueuedURIMetaData,
            "DesiredFirstTrackNumberEnqueued": DesiredFirstTrackNumberEnqueued,
            "EnqueueAsNext": EnqueueAsNext,
        }

        out_params = self.call_action("AddURIToQueue", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("FirstTrackNumberEnqueued", "NumTracksAdded", "NewQueueLength",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_AddURIToSavedQueue(self, InstanceID, ObjectID, UpdateID, EnqueuedURI, EnqueuedURIMetaData, AddAtIndex, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AddURIToSavedQueue action.

            :returns: "NumTracksAdded", "NewQueueLength", "NewUpdateID"
        """
        arguments = {
            "InstanceID": InstanceID,
            "ObjectID": ObjectID,
            "UpdateID": UpdateID,
            "EnqueuedURI": EnqueuedURI,
            "EnqueuedURIMetaData": EnqueuedURIMetaData,
            "AddAtIndex": AddAtIndex,
        }

        out_params = self.call_action("AddURIToSavedQueue", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NumTracksAdded", "NewQueueLength", "NewUpdateID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_BackupQueue(self, InstanceID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the BackupQueue action.
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        self.call_action("BackupQueue", arguments=arguments, aspects=aspects)

        return

    def action_BecomeCoordinatorOfStandaloneGroup(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the BecomeCoordinatorOfStandaloneGroup action.

            :returns: "DelegatedGroupCoordinatorID", "NewGroupID"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("BecomeCoordinatorOfStandaloneGroup", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("DelegatedGroupCoordinatorID", "NewGroupID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_BecomeGroupCoordinator(self, InstanceID, CurrentCoordinator, CurrentGroupID, OtherMembers, TransportSettings, CurrentURI, CurrentURIMetaData, SleepTimerState, AlarmState, StreamRestartState, CurrentQueueTrackList, CurrentVLIState, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the BecomeGroupCoordinator action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "CurrentCoordinator": CurrentCoordinator,
            "CurrentGroupID": CurrentGroupID,
            "OtherMembers": OtherMembers,
            "TransportSettings": TransportSettings,
            "CurrentURI": CurrentURI,
            "CurrentURIMetaData": CurrentURIMetaData,
            "SleepTimerState": SleepTimerState,
            "AlarmState": AlarmState,
            "StreamRestartState": StreamRestartState,
            "CurrentQueueTrackList": CurrentQueueTrackList,
            "CurrentVLIState": CurrentVLIState,
        }

        self.call_action("BecomeGroupCoordinator", arguments=arguments, aspects=aspects)

        return

    def action_BecomeGroupCoordinatorAndSource(self, InstanceID, CurrentCoordinator, CurrentGroupID, OtherMembers, CurrentURI, CurrentURIMetaData, SleepTimerState, AlarmState, StreamRestartState, CurrentAVTTrackList, CurrentQueueTrackList, CurrentSourceState, ResumePlayback, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the BecomeGroupCoordinatorAndSource action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "CurrentCoordinator": CurrentCoordinator,
            "CurrentGroupID": CurrentGroupID,
            "OtherMembers": OtherMembers,
            "CurrentURI": CurrentURI,
            "CurrentURIMetaData": CurrentURIMetaData,
            "SleepTimerState": SleepTimerState,
            "AlarmState": AlarmState,
            "StreamRestartState": StreamRestartState,
            "CurrentAVTTrackList": CurrentAVTTrackList,
            "CurrentQueueTrackList": CurrentQueueTrackList,
            "CurrentSourceState": CurrentSourceState,
            "ResumePlayback": ResumePlayback,
        }

        self.call_action("BecomeGroupCoordinatorAndSource", arguments=arguments, aspects=aspects)

        return

    def action_ChangeCoordinator(self, InstanceID, CurrentCoordinator, NewCoordinator, NewTransportSettings, CurrentAVTransportURI, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ChangeCoordinator action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "CurrentCoordinator": CurrentCoordinator,
            "NewCoordinator": NewCoordinator,
            "NewTransportSettings": NewTransportSettings,
            "CurrentAVTransportURI": CurrentAVTransportURI,
        }

        self.call_action("ChangeCoordinator", arguments=arguments, aspects=aspects)

        return

    def action_ChangeTransportSettings(self, InstanceID, NewTransportSettings, CurrentAVTransportURI, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ChangeTransportSettings action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "NewTransportSettings": NewTransportSettings,
            "CurrentAVTransportURI": CurrentAVTransportURI,
        }

        self.call_action("ChangeTransportSettings", arguments=arguments, aspects=aspects)

        return

    def action_ConfigureSleepTimer(self, InstanceID, NewSleepTimerDuration, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ConfigureSleepTimer action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "NewSleepTimerDuration": NewSleepTimerDuration,
        }

        self.call_action("ConfigureSleepTimer", arguments=arguments, aspects=aspects)

        return

    def action_CreateSavedQueue(self, InstanceID, Title, EnqueuedURI, EnqueuedURIMetaData, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the CreateSavedQueue action.

            :returns: "NumTracksAdded", "NewQueueLength", "AssignedObjectID", "NewUpdateID"
        """
        arguments = {
            "InstanceID": InstanceID,
            "Title": Title,
            "EnqueuedURI": EnqueuedURI,
            "EnqueuedURIMetaData": EnqueuedURIMetaData,
        }

        out_params = self.call_action("CreateSavedQueue", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NumTracksAdded", "NewQueueLength", "AssignedObjectID", "NewUpdateID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_DelegateGroupCoordinationTo(self, InstanceID, NewCoordinator, RejoinGroup, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DelegateGroupCoordinationTo action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "NewCoordinator": NewCoordinator,
            "RejoinGroup": RejoinGroup,
        }

        self.call_action("DelegateGroupCoordinationTo", arguments=arguments, aspects=aspects)

        return

    def action_EndDirectControlSession(self, InstanceID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the EndDirectControlSession action.
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        self.call_action("EndDirectControlSession", arguments=arguments, aspects=aspects)

        return

    def action_GetCrossfadeMode(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetCrossfadeMode action.

            :returns: "CrossfadeMode"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetCrossfadeMode", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CrossfadeMode",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetCurrentTransportActions(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetCurrentTransportActions action.

            :returns: "Actions"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetCurrentTransportActions", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Actions",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDeviceCapabilities(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDeviceCapabilities action.

            :returns: "PlayMedia", "RecMedia", "RecQualityModes"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetDeviceCapabilities", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("PlayMedia", "RecMedia", "RecQualityModes",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetMediaInfo(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetMediaInfo action.

            :returns: "NrTracks", "MediaDuration", "CurrentURI", "CurrentURIMetaData", "NextURI", "NextURIMetaData", "PlayMedium", "RecordMedium", "WriteStatus"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetMediaInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NrTracks", "MediaDuration", "CurrentURI", "CurrentURIMetaData", "NextURI", "NextURIMetaData", "PlayMedium", "RecordMedium", "WriteStatus",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetPositionInfo(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetPositionInfo action.

            :returns: "Track", "TrackDuration", "TrackMetaData", "TrackURI", "RelTime", "AbsTime", "RelCount", "AbsCount"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetPositionInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Track", "TrackDuration", "TrackMetaData", "TrackURI", "RelTime", "AbsTime", "RelCount", "AbsCount",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetRemainingSleepTimerDuration(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetRemainingSleepTimerDuration action.

            :returns: "RemainingSleepTimerDuration", "CurrentSleepTimerGeneration"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetRemainingSleepTimerDuration", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("RemainingSleepTimerDuration", "CurrentSleepTimerGeneration",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetRunningAlarmProperties(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetRunningAlarmProperties action.

            :returns: "AlarmID", "GroupID", "LoggedStartTime"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetRunningAlarmProperties", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("AlarmID", "GroupID", "LoggedStartTime",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTransportInfo(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTransportInfo action.

            :returns: "CurrentTransportState", "CurrentTransportStatus", "CurrentSpeed"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetTransportInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentTransportState", "CurrentTransportStatus", "CurrentSpeed",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTransportSettings(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTransportSettings action.

            :returns: "PlayMode", "RecQualityMode"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetTransportSettings", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("PlayMode", "RecQualityMode",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_Next(self, InstanceID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Next action.
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        self.call_action("Next", arguments=arguments, aspects=aspects)

        return

    def action_NotifyDeletedURI(self, InstanceID, DeletedURI, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the NotifyDeletedURI action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "DeletedURI": DeletedURI,
        }

        self.call_action("NotifyDeletedURI", arguments=arguments, aspects=aspects)

        return

    def action_Pause(self, InstanceID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Pause action.
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        self.call_action("Pause", arguments=arguments, aspects=aspects)

        return

    def action_Play(self, InstanceID, Speed, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Play action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "Speed": Speed,
        }

        self.call_action("Play", arguments=arguments, aspects=aspects)

        return

    def action_Previous(self, InstanceID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Previous action.
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        self.call_action("Previous", arguments=arguments, aspects=aspects)

        return

    def action_RemoveAllTracksFromQueue(self, InstanceID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RemoveAllTracksFromQueue action.
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        self.call_action("RemoveAllTracksFromQueue", arguments=arguments, aspects=aspects)

        return

    def action_RemoveTrackFromQueue(self, InstanceID, ObjectID, UpdateID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RemoveTrackFromQueue action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "ObjectID": ObjectID,
            "UpdateID": UpdateID,
        }

        self.call_action("RemoveTrackFromQueue", arguments=arguments, aspects=aspects)

        return

    def action_RemoveTrackRangeFromQueue(self, InstanceID, UpdateID, StartingIndex, NumberOfTracks, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RemoveTrackRangeFromQueue action.

            :returns: "NewUpdateID"
        """
        arguments = {
            "InstanceID": InstanceID,
            "UpdateID": UpdateID,
            "StartingIndex": StartingIndex,
            "NumberOfTracks": NumberOfTracks,
        }

        out_params = self.call_action("RemoveTrackRangeFromQueue", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewUpdateID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_ReorderTracksInQueue(self, InstanceID, StartingIndex, NumberOfTracks, InsertBefore, UpdateID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ReorderTracksInQueue action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "StartingIndex": StartingIndex,
            "NumberOfTracks": NumberOfTracks,
            "InsertBefore": InsertBefore,
            "UpdateID": UpdateID,
        }

        self.call_action("ReorderTracksInQueue", arguments=arguments, aspects=aspects)

        return

    def action_ReorderTracksInSavedQueue(self, InstanceID, ObjectID, UpdateID, TrackList, NewPositionList, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ReorderTracksInSavedQueue action.

            :returns: "QueueLengthChange", "NewQueueLength", "NewUpdateID"
        """
        arguments = {
            "InstanceID": InstanceID,
            "ObjectID": ObjectID,
            "UpdateID": UpdateID,
            "TrackList": TrackList,
            "NewPositionList": NewPositionList,
        }

        out_params = self.call_action("ReorderTracksInSavedQueue", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("QueueLengthChange", "NewQueueLength", "NewUpdateID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_RunAlarm(self, InstanceID, AlarmID, LoggedStartTime, Duration, ProgramURI, ProgramMetaData, PlayMode, Volume, IncludeLinkedZones, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RunAlarm action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "AlarmID": AlarmID,
            "LoggedStartTime": LoggedStartTime,
            "Duration": Duration,
            "ProgramURI": ProgramURI,
            "ProgramMetaData": ProgramMetaData,
            "PlayMode": PlayMode,
            "Volume": Volume,
            "IncludeLinkedZones": IncludeLinkedZones,
        }

        self.call_action("RunAlarm", arguments=arguments, aspects=aspects)

        return

    def action_SaveQueue(self, InstanceID, Title, ObjectID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SaveQueue action.

            :returns: "AssignedObjectID"
        """
        arguments = {
            "InstanceID": InstanceID,
            "Title": Title,
            "ObjectID": ObjectID,
        }

        out_params = self.call_action("SaveQueue", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("AssignedObjectID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_Seek(self, InstanceID, Unit, Target, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Seek action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "Unit": Unit,
            "Target": Target,
        }

        self.call_action("Seek", arguments=arguments, aspects=aspects)

        return

    def action_SetAVTransportURI(self, InstanceID, CurrentURI, CurrentURIMetaData, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetAVTransportURI action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "CurrentURI": CurrentURI,
            "CurrentURIMetaData": CurrentURIMetaData,
        }

        self.call_action("SetAVTransportURI", arguments=arguments, aspects=aspects)

        return

    def action_SetCrossfadeMode(self, InstanceID, CrossfadeMode, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetCrossfadeMode action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "CrossfadeMode": CrossfadeMode,
        }

        self.call_action("SetCrossfadeMode", arguments=arguments, aspects=aspects)

        return

    def action_SetNextAVTransportURI(self, InstanceID, NextURI, NextURIMetaData, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetNextAVTransportURI action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "NextURI": NextURI,
            "NextURIMetaData": NextURIMetaData,
        }

        self.call_action("SetNextAVTransportURI", arguments=arguments, aspects=aspects)

        return

    def action_SetPlayMode(self, InstanceID, NewPlayMode, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetPlayMode action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "NewPlayMode": NewPlayMode,
        }

        self.call_action("SetPlayMode", arguments=arguments, aspects=aspects)

        return

    def action_SnoozeAlarm(self, InstanceID, Duration, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SnoozeAlarm action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "Duration": Duration,
        }

        self.call_action("SnoozeAlarm", arguments=arguments, aspects=aspects)

        return

    def action_StartAutoplay(self, InstanceID, ProgramURI, ProgramMetaData, Volume, IncludeLinkedZones, ResetVolumeAfter, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the StartAutoplay action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "ProgramURI": ProgramURI,
            "ProgramMetaData": ProgramMetaData,
            "Volume": Volume,
            "IncludeLinkedZones": IncludeLinkedZones,
            "ResetVolumeAfter": ResetVolumeAfter,
        }

        self.call_action("StartAutoplay", arguments=arguments, aspects=aspects)

        return

    def action_Stop(self, InstanceID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Stop action.
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        self.call_action("Stop", arguments=arguments, aspects=aspects)

        return



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class Feeder1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:Feeder:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'Feeder1'

    SERVICE_DEFAULT_VARIABLES = {
        "EntireDocument": { "data_type": "string", "default": "1", "allowed_list": "['1', '0', 'device-setting']"},
        "FailureCode": { "data_type": "string", "default": "None", "allowed_list": "['None', 'Jammed', 'Timeout']"},
        "FeederMode": { "data_type": "string", "default": "Simplex", "allowed_list": "['Simplex']"},
        "InputJustification": { "data_type": "string", "default": None, "allowed_list": None},
        "JobID": { "data_type": "ui4", "default": "0", "allowed_list": None},
        "Model": { "data_type": "string", "default": None, "allowed_list": None},
        "SheetHeight": { "data_type": "ui4", "default": None, "allowed_list": None},
        "SheetWidth": { "data_type": "ui4", "default": None, "allowed_list": None},
        "State": { "data_type": "string", "default": "Unloaded", "allowed_list": "['Unloaded', 'Loaded', 'Busy', 'Erred']"},
        "Timeout": { "data_type": "ui4", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {
        "MorePages": { "data_type": "boolean", "default": "0", "allowed_list": None},
    }

    def action_Eject(self, JobIDIn, EntireDocumentIn, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Eject action.

            :returns: "StateOut"
        """
        arguments = {
            "JobIDIn": JobIDIn,
            "EntireDocumentIn": EntireDocumentIn,
        }

        out_params = self.call_action("Eject", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("StateOut",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetFeederMode(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetFeederMode action.

            :returns: "FeederModeOut"
        """
        arguments = { }

        out_params = self.call_action("GetFeederMode", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("FeederModeOut",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetState(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetState action.

            :returns: "StateOut", "MorePagesOut", "FailureCodeOut"
        """
        arguments = { }

        out_params = self.call_action("GetState", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("StateOut", "MorePagesOut", "FailureCodeOut",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_Load(self, JobIDIn, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Load action.

            :returns: "StateOut"
        """
        arguments = {
            "JobIDIn": JobIDIn,
        }

        out_params = self.call_action("Load", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("StateOut",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_Reset(self, JobIDIn, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Reset action.

            :returns: "StateOut"
        """
        arguments = {
            "JobIDIn": JobIDIn,
        }

        out_params = self.call_action("Reset", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("StateOut",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetFeederMode(self, JobIDIn, FeederModeIn, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetFeederMode action.
        """
        arguments = {
            "JobIDIn": JobIDIn,
            "FeederModeIn": FeederModeIn,
        }

        self.call_action("SetFeederMode", arguments=arguments, aspects=aspects)

        return



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class LANHostConfigManagement1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:LANHostConfigManagement:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'LANHostConfigManagement1'

    SERVICE_DEFAULT_VARIABLES = {
        "DHCPRelay": { "data_type": "boolean", "default": None, "allowed_list": None},
        "DHCPServerConfigurable": { "data_type": "boolean", "default": None, "allowed_list": None},
        "DNSServers": { "data_type": "string", "default": None, "allowed_list": None},
        "DomainName": { "data_type": "string", "default": None, "allowed_list": None},
        "IPRouters": { "data_type": "string", "default": None, "allowed_list": None},
        "MaxAddress": { "data_type": "string", "default": None, "allowed_list": None},
        "MinAddress": { "data_type": "string", "default": None, "allowed_list": None},
        "ReservedAddresses": { "data_type": "string", "default": None, "allowed_list": None},
        "SubnetMask": { "data_type": "string", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {}

    def action_DeleteDNSServer(self, NewDNSServers, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DeleteDNSServer action.
        """
        arguments = {
            "NewDNSServers": NewDNSServers,
        }

        self.call_action("DeleteDNSServer", arguments=arguments, aspects=aspects)

        return

    def action_DeleteIPRouter(self, NewIPRouters, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DeleteIPRouter action.
        """
        arguments = {
            "NewIPRouters": NewIPRouters,
        }

        self.call_action("DeleteIPRouter", arguments=arguments, aspects=aspects)

        return

    def action_DeleteReservedAddress(self, NewReservedAddresses, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DeleteReservedAddress action.
        """
        arguments = {
            "NewReservedAddresses": NewReservedAddresses,
        }

        self.call_action("DeleteReservedAddress", arguments=arguments, aspects=aspects)

        return

    def action_GetAddressRange(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetAddressRange action.

            :returns: "NewMinAddress", "NewMaxAddress"
        """
        arguments = { }

        out_params = self.call_action("GetAddressRange", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewMinAddress", "NewMaxAddress",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDHCPRelay(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDHCPRelay action.

            :returns: "NewDHCPRelay"
        """
        arguments = { }

        out_params = self.call_action("GetDHCPRelay", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewDHCPRelay",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDHCPServerConfigurable(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDHCPServerConfigurable action.

            :returns: "NewDHCPServerConfigurable"
        """
        arguments = { }

        out_params = self.call_action("GetDHCPServerConfigurable", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewDHCPServerConfigurable",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDNSServers(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDNSServers action.

            :returns: "NewDNSServers"
        """
        arguments = { }

        out_params = self.call_action("GetDNSServers", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewDNSServers",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDomainName(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDomainName action.

            :returns: "NewDomainName"
        """
        arguments = { }

        out_params = self.call_action("GetDomainName", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewDomainName",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetIPRoutersList(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetIPRoutersList action.

            :returns: "NewIPRouters"
        """
        arguments = { }

        out_params = self.call_action("GetIPRoutersList", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewIPRouters",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetReservedAddresses(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetReservedAddresses action.

            :returns: "NewReservedAddresses"
        """
        arguments = { }

        out_params = self.call_action("GetReservedAddresses", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewReservedAddresses",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSubnetMask(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSubnetMask action.

            :returns: "NewSubnetMask"
        """
        arguments = { }

        out_params = self.call_action("GetSubnetMask", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewSubnetMask",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetAddressRange(self, NewMinAddress, NewMaxAddress, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetAddressRange action.
        """
        arguments = {
            "NewMinAddress": NewMinAddress,
            "NewMaxAddress": NewMaxAddress,
        }

        self.call_action("SetAddressRange", arguments=arguments, aspects=aspects)

        return

    def action_SetDHCPRelay(self, NewDHCPRelay, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetDHCPRelay action.
        """
        arguments = {
            "NewDHCPRelay": NewDHCPRelay,
        }

        self.call_action("SetDHCPRelay", arguments=arguments, aspects=aspects)

        return

    def action_SetDHCPServerConfigurable(self, NewDHCPServerConfigurable, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetDHCPServerConfigurable action.
        """
        arguments = {
            "NewDHCPServerConfigurable": NewDHCPServerConfigurable,
        }

        self.call_action("SetDHCPServerConfigurable", arguments=arguments, aspects=aspects)

        return

    def action_SetDNSServer(self, NewDNSServers, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetDNSServer action.
        """
        arguments = {
            "NewDNSServers": NewDNSServers,
        }

        self.call_action("SetDNSServer", arguments=arguments, aspects=aspects)

        return

    def action_SetDomainName(self, NewDomainName, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetDomainName action.
        """
        arguments = {
            "NewDomainName": NewDomainName,
        }

        self.call_action("SetDomainName", arguments=arguments, aspects=aspects)

        return

    def action_SetIPRouter(self, NewIPRouters, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetIPRouter action.
        """
        arguments = {
            "NewIPRouters": NewIPRouters,
        }

        self.call_action("SetIPRouter", arguments=arguments, aspects=aspects)

        return

    def action_SetReservedAddress(self, NewReservedAddresses, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetReservedAddress action.
        """
        arguments = {
            "NewReservedAddresses": NewReservedAddresses,
        }

        self.call_action("SetReservedAddress", arguments=arguments, aspects=aspects)

        return

    def action_SetSubnetMask(self, NewSubnetMask, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetSubnetMask action.
        """
        arguments = {
            "NewSubnetMask": NewSubnetMask,
        }

        self.call_action("SetSubnetMask", arguments=arguments, aspects=aspects)

        return



"""
.. module:: metadata
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module that contains decorators that can be utilized to attach metadata to automation
               tasks, tests, scopes, and integrations.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

# pylint: disable=too-few-public-methods

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

class Category:
    """
        The 'Category' decorator allows test engineers to associate categories with individual test
        cases or collections of tests.
    """

    def __init__(self, *categories):
        if len(categories) == 0:
            raise ValueError("The 'Category' decorator requires at least one category arguement.")
        self._categories = categories
        return

    def __call__(self, obj_to_decorate):
        if not hasattr(obj_to_decorate, "_metadata_"):
            obj_to_decorate._metadata_ = {}
        if "categories" in obj_to_decorate._metadata_:
            obj_to_decorate._metadata_["categories"].extend(self._categories)
        else:
            obj_to_decorate._metadata_["categories"] = self._categories
        return obj_to_decorate

class Keywords:
    """
        The 'Keywords' decorator allows test engineers to associate keywords with individual test
        cases or collections of tests.
    """

    def __init__(self, *keywords):
        if len(keywords) == 0:
            raise ValueError("The 'Keyword' decorator requires at least one keyword arguement.")
        self._keywords = keywords
        return

    def __call__(self, obj_to_decorate):
        if not hasattr(obj_to_decorate, "_metadata_"):
            obj_to_decorate._metadata_ = {}
        if "keywords" in obj_to_decorate._metadata_:
            obj_to_decorate._metadata_["keywords"].extend(self._keywords)
        else:
            obj_to_decorate._metadata_["keywords"] = self._keywords
        return obj_to_decorate

class Priority:
    """
        The 'Priority' decorator allows test engineers to associate a test priority with
        individual test cases or collections of tests.
    """

    def __init__(self, priority):
        self._priority = priority
        return

    def __call__(self, obj_to_decorate):
        if not hasattr(obj_to_decorate, "_metadata_"):
            obj_to_decorate._metadata_ = {}
        if "priority" in obj_to_decorate._metadata_:
            obj_to_decorate._metadata_["priority"].extend(self._priority)
        else:
            obj_to_decorate._metadata_["priority"] = self._priority
        return obj_to_decorate




__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import List

import os

from sqlalchemy import create_engine
from sqlalchemy.engine.base import Engine as BaseEngine

from akit.environment.contextpaths import ContextPaths
from akit.exceptions import AKitConfigurationError, AKitNotOverloadedError
from akit.interop.credentials.basiccredential import BasicCredential
from akit.interop.credentials.credentialmanager import CredentialManager

from akit.xlogging.foundations import getAutomatonKitLogger

logger = getAutomatonKitLogger()

class DatabaseConnectionFactory:
    def create_engine(self, *, dbname=None, echo=True) -> BaseEngine:
        errmsg = "The 'create_engine' method must be overloaded by 'DatabaseConnectionFactory' derived types."
        raise AKitNotOverloadedError(errmsg)

class BasicDatabaseConnectionFactory(DatabaseConnectionFactory):

    def __init__(self, profile_name, *, conntype: str, dbtype:str, credential: str, dbname: str=None):
        self.profile_name = profile_name
        self.conntype = conntype
        self.dbtype = dbtype
        self.dbname = dbname
        self.credential = credential
        return

    def create_engine(self, *, dbname=None, echo=True) -> BaseEngine:
        # TODO: Implement pipe based DB connection engine creation
        return

class BasicTcpDatabaseConnectionFactory(DatabaseConnectionFactory):

    def __init__(self, profile_name, *, conntype: str, dbtype:str, host: str, port: int, credential: str, dbname: str=None):
        self.profile_name = profile_name
        self.conntype = conntype
        self.dbtype = dbtype
        self.host = host
        self.port = port
        self.credential = credential
        self.dbname = dbname
        return

    def create_engine(self, *, dbname=None, echo=True) -> BaseEngine:

        # If the database type we are creating an engine for is postgresql, make sure
        # we can import all of the postgresql dependencies.
        if self.dbtype == 'postgresql':
            try:
                import psycopg2
            except ImportError:
                errmsg_lines = [
                    "In order to create a postgresql database engine, you must install the following dependencies.",
                    "DEPENDENCIES",
                    "    psycopg2"
                ]
                errmsg = os.linesep.join(errmsg_lines)
                raise AKitConfigurationError(errmsg)

        if dbname is None and self.dbname is not None:
            dbname = self.dbname

        if dbname is None:
            warnmsg = "BasicTcpDatabaseConnectionFactory.create_engine: dbname was None."
            logger.warn(warnmsg)

        credmgr = CredentialManager()

        cred = credmgr.credentials[self.credential]
        if isinstance(cred, BasicCredential):
            connstr = '%s+psycopg2://%s:%s@%s:%d/%s' % (
                self.dbtype, cred.username, cred.password, self.host, self.port, dbname)
            dbengine = create_engine(connstr, echo=echo)

        return dbengine

database_connection_factories = {}

def lookup_database_connection_factory(conn_profile: str):

    global database_connection_factories

    conn_factory = None

    if conn_profile in database_connection_factories:
        conn_factory = database_connection_factories[conn_profile]
    else:
        from akit.environment.context import Context

        ctx = Context()
        
        conn_info = None

        rcdatabases = ctx.lookup(ContextPaths.DATABASES)
        if rcdatabases is not None and conn_profile in rcdatabases:
            conn_info = rcdatabases[conn_profile].value
        else:
            from akit.wellknown.singletons import LandscapeSingleton

            lscape = LandscapeSingleton()
            lsdatabases = lscape.databases
            if lsdatabases is not None and conn_profile in lsdatabases:
                conn_info = lsdatabases[conn_profile]

        if conn_info is not None:
            if "conntype" in conn_info:
                conntype = conn_info["conntype"].lower()
                if conntype == "basic":
                    conn_factory = BasicDatabaseConnectionFactory(conn_profile, **conn_info)
                elif conntype == "basic-tcp":
                    conn_factory = BasicTcpDatabaseConnectionFactory(conn_profile, **conn_info)
                else:
                    errmsg = "Unknown database connection type. connection={} conntype={}".format(
                        conn_profile, conntype)
                    raise AKitConfigurationError(errmsg)

                database_connection_factories[conn_profile] = conn_factory
            else:
                errmsg = "Database connection entries must have a 'conntype' entry. connection={}".format(
                    conn_profile
                )
                raise AKitConfigurationError(errmsg)
        else:
            errmsg = "Database connection not found. connection={}".format(
                conn_profile
            )
            raise AKitConfigurationError(errmsg)

    return conn_factory


"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class DeviceProperties1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'DeviceProperties1' service.
    """

    SERVICE_MANUFACTURER = 'SonosInc'
    SERVICE_TYPE = 'urn:schemas-upnp-org:service:DeviceProperties:1'

    SERVICE_DEFAULT_VARIABLES = {
        "AutoplayIncludeLinkedZones": { "data_type": "boolean", "default": None, "allowed_list": None},
        "AutoplayRoomUUID": { "data_type": "string", "default": None, "allowed_list": None},
        "AutoplaySource": { "data_type": "string", "default": None, "allowed_list": None},
        "AutoplayUseVolume": { "data_type": "boolean", "default": None, "allowed_list": None},
        "AutoplayVolume": { "data_type": "ui2", "default": None, "allowed_list": None},
        "ButtonLockState": { "data_type": "string", "default": None, "allowed_list": "['On', 'Off']"},
        "CopyrightInfo": { "data_type": "string", "default": None, "allowed_list": None},
        "DisplaySoftwareVersion": { "data_type": "string", "default": None, "allowed_list": None},
        "ExtraInfo": { "data_type": "string", "default": None, "allowed_list": None},
        "Flags": { "data_type": "ui4", "default": None, "allowed_list": None},
        "HTAudioIn": { "data_type": "ui4", "default": None, "allowed_list": None},
        "HardwareVersion": { "data_type": "string", "default": None, "allowed_list": None},
        "HouseholdID": { "data_type": "string", "default": None, "allowed_list": None},
        "IPAddress": { "data_type": "string", "default": None, "allowed_list": None},
        "KeepGrouped": { "data_type": "boolean", "default": None, "allowed_list": None},
        "LEDState": { "data_type": "string", "default": None, "allowed_list": "['On', 'Off']"},
        "MACAddress": { "data_type": "string", "default": None, "allowed_list": None},
        "SatRoomUUID": { "data_type": "string", "default": None, "allowed_list": None},
        "SerialNumber": { "data_type": "string", "default": None, "allowed_list": None},
        "SoftwareVersion": { "data_type": "string", "default": None, "allowed_list": None},
        "TargetRoomName": { "data_type": "string", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {
        "AirPlayEnabled": { "data_type": "boolean", "default": None, "allowed_list": None},
        "AvailableRoomCalibration": { "data_type": "string", "default": None, "allowed_list": None},
        "BehindWifiExtender": { "data_type": "ui4", "default": None, "allowed_list": None},
        "ChannelFreq": { "data_type": "ui4", "default": None, "allowed_list": None},
        "ChannelMapSet": { "data_type": "string", "default": None, "allowed_list": None},
        "ConfigMode": { "data_type": "string", "default": None, "allowed_list": None},
        "Configuration": { "data_type": "string", "default": None, "allowed_list": None},
        "HTBondedZoneCommitState": { "data_type": "ui4", "default": None, "allowed_list": None},
        "HTFreq": { "data_type": "ui4", "default": None, "allowed_list": None},
        "HTSatChanMapSet": { "data_type": "string", "default": None, "allowed_list": None},
        "HasConfiguredSSID": { "data_type": "boolean", "default": None, "allowed_list": None},
        "HdmiCecAvailable": { "data_type": "boolean", "default": None, "allowed_list": None},
        "Icon": { "data_type": "string", "default": None, "allowed_list": None},
        "Invisible": { "data_type": "boolean", "default": None, "allowed_list": None},
        "IsIdle": { "data_type": "boolean", "default": None, "allowed_list": None},
        "IsZoneBridge": { "data_type": "boolean", "default": None, "allowed_list": None},
        "LastChangedPlayState": { "data_type": "string", "default": None, "allowed_list": None},
        "MicEnabled": { "data_type": "ui4", "default": None, "allowed_list": None},
        "MoreInfo": { "data_type": "string", "default": None, "allowed_list": None},
        "Orientation": { "data_type": "i4", "default": None, "allowed_list": None},
        "RoomCalibrationState": { "data_type": "i4", "default": None, "allowed_list": None},
        "SecureRegState": { "data_type": "ui4", "default": None, "allowed_list": None},
        "SettingsReplicationState": { "data_type": "string", "default": None, "allowed_list": None},
        "SupportsAudioClip": { "data_type": "boolean", "default": None, "allowed_list": None},
        "SupportsAudioIn": { "data_type": "boolean", "default": None, "allowed_list": None},
        "TVConfigurationError": { "data_type": "boolean", "default": None, "allowed_list": None},
        "VoiceConfigState": { "data_type": "ui4", "default": None, "allowed_list": None},
        "WifiEnabled": { "data_type": "boolean", "default": None, "allowed_list": None},
        "WirelessLeafOnly": { "data_type": "boolean", "default": None, "allowed_list": None},
        "WirelessMode": { "data_type": "ui4", "default": None, "allowed_list": None},
        "ZoneName": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_AddBondedZones(self, ChannelMapSet, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AddBondedZones action.
        """
        arguments = {
            "ChannelMapSet": ChannelMapSet,
        }

        self.call_action("AddBondedZones", arguments=arguments, aspects=aspects)

        return

    def action_AddHTSatellite(self, HTSatChanMapSet, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AddHTSatellite action.
        """
        arguments = {
            "HTSatChanMapSet": HTSatChanMapSet,
        }

        self.call_action("AddHTSatellite", arguments=arguments, aspects=aspects)

        return

    def action_CreateStereoPair(self, ChannelMapSet, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the CreateStereoPair action.
        """
        arguments = {
            "ChannelMapSet": ChannelMapSet,
        }

        self.call_action("CreateStereoPair", arguments=arguments, aspects=aspects)

        return

    def action_EnterConfigMode(self, Mode, Options, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the EnterConfigMode action.

            :returns: "State"
        """
        arguments = {
            "Mode": Mode,
            "Options": Options,
        }

        out_params = self.call_action("EnterConfigMode", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("State",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_ExitConfigMode(self, Options, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ExitConfigMode action.
        """
        arguments = {
            "Options": Options,
        }

        self.call_action("ExitConfigMode", arguments=arguments, aspects=aspects)

        return

    def action_GetAutoplayLinkedZones(self, Source, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetAutoplayLinkedZones action.

            :returns: "IncludeLinkedZones"
        """
        arguments = {
            "Source": Source,
        }

        out_params = self.call_action("GetAutoplayLinkedZones", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("IncludeLinkedZones",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetAutoplayRoomUUID(self, Source, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetAutoplayRoomUUID action.

            :returns: "RoomUUID"
        """
        arguments = {
            "Source": Source,
        }

        out_params = self.call_action("GetAutoplayRoomUUID", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("RoomUUID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetAutoplayVolume(self, Source, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetAutoplayVolume action.

            :returns: "CurrentVolume"
        """
        arguments = {
            "Source": Source,
        }

        out_params = self.call_action("GetAutoplayVolume", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentVolume",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetButtonLockState(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetButtonLockState action.

            :returns: "CurrentButtonLockState"
        """
        arguments = { }

        out_params = self.call_action("GetButtonLockState", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentButtonLockState",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetButtonState(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetButtonState action.

            :returns: "State"
        """
        arguments = { }

        out_params = self.call_action("GetButtonState", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("State",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetHouseholdID(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetHouseholdID action.

            :returns: "CurrentHouseholdID"
        """
        arguments = { }

        out_params = self.call_action("GetHouseholdID", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentHouseholdID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetLEDState(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetLEDState action.

            :returns: "CurrentLEDState"
        """
        arguments = { }

        out_params = self.call_action("GetLEDState", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentLEDState",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetUseAutoplayVolume(self, Source, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetUseAutoplayVolume action.

            :returns: "UseVolume"
        """
        arguments = {
            "Source": Source,
        }

        out_params = self.call_action("GetUseAutoplayVolume", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("UseVolume",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetZoneAttributes(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetZoneAttributes action.

            :returns: "CurrentZoneName", "CurrentIcon", "CurrentConfiguration", "CurrentTargetRoomName"
        """
        arguments = { }

        out_params = self.call_action("GetZoneAttributes", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentZoneName", "CurrentIcon", "CurrentConfiguration", "CurrentTargetRoomName",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetZoneInfo(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetZoneInfo action.

            :returns: "SerialNumber", "SoftwareVersion", "DisplaySoftwareVersion", "HardwareVersion", "IPAddress", "MACAddress", "CopyrightInfo", "ExtraInfo", "HTAudioIn", "Flags"
        """
        arguments = { }

        out_params = self.call_action("GetZoneInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("SerialNumber", "SoftwareVersion", "DisplaySoftwareVersion", "HardwareVersion", "IPAddress", "MACAddress", "CopyrightInfo", "ExtraInfo", "HTAudioIn", "Flags",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_RemoveBondedZones(self, ChannelMapSet, KeepGrouped, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RemoveBondedZones action.
        """
        arguments = {
            "ChannelMapSet": ChannelMapSet,
            "KeepGrouped": KeepGrouped,
        }

        self.call_action("RemoveBondedZones", arguments=arguments, aspects=aspects)

        return

    def action_RemoveHTSatellite(self, SatRoomUUID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RemoveHTSatellite action.
        """
        arguments = {
            "SatRoomUUID": SatRoomUUID,
        }

        self.call_action("RemoveHTSatellite", arguments=arguments, aspects=aspects)

        return

    def action_RoomDetectionStartChirping(self, Channel, DurationMilliseconds, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RoomDetectionStartChirping action.

            :returns: "PlayId", "ChirpIfPlayingSwappableAudio"
        """
        arguments = {
            "Channel": Channel,
            "DurationMilliseconds": DurationMilliseconds,
        }

        out_params = self.call_action("RoomDetectionStartChirping", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("PlayId", "ChirpIfPlayingSwappableAudio",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_RoomDetectionStopChirping(self, PlayId, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RoomDetectionStopChirping action.
        """
        arguments = {
            "PlayId": PlayId,
        }

        self.call_action("RoomDetectionStopChirping", arguments=arguments, aspects=aspects)

        return

    def action_SeparateStereoPair(self, ChannelMapSet, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SeparateStereoPair action.
        """
        arguments = {
            "ChannelMapSet": ChannelMapSet,
        }

        self.call_action("SeparateStereoPair", arguments=arguments, aspects=aspects)

        return

    def action_SetAutoplayLinkedZones(self, IncludeLinkedZones, Source, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetAutoplayLinkedZones action.
        """
        arguments = {
            "IncludeLinkedZones": IncludeLinkedZones,
            "Source": Source,
        }

        self.call_action("SetAutoplayLinkedZones", arguments=arguments, aspects=aspects)

        return

    def action_SetAutoplayRoomUUID(self, RoomUUID, Source, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetAutoplayRoomUUID action.
        """
        arguments = {
            "RoomUUID": RoomUUID,
            "Source": Source,
        }

        self.call_action("SetAutoplayRoomUUID", arguments=arguments, aspects=aspects)

        return

    def action_SetAutoplayVolume(self, Volume, Source, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetAutoplayVolume action.
        """
        arguments = {
            "Volume": Volume,
            "Source": Source,
        }

        self.call_action("SetAutoplayVolume", arguments=arguments, aspects=aspects)

        return

    def action_SetButtonLockState(self, DesiredButtonLockState, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetButtonLockState action.
        """
        arguments = {
            "DesiredButtonLockState": DesiredButtonLockState,
        }

        self.call_action("SetButtonLockState", arguments=arguments, aspects=aspects)

        return

    def action_SetLEDState(self, DesiredLEDState, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetLEDState action.
        """
        arguments = {
            "DesiredLEDState": DesiredLEDState,
        }

        self.call_action("SetLEDState", arguments=arguments, aspects=aspects)

        return

    def action_SetUseAutoplayVolume(self, UseVolume, Source, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetUseAutoplayVolume action.
        """
        arguments = {
            "UseVolume": UseVolume,
            "Source": Source,
        }

        self.call_action("SetUseAutoplayVolume", arguments=arguments, aspects=aspects)

        return

    def action_SetZoneAttributes(self, DesiredZoneName, DesiredIcon, DesiredConfiguration, DesiredTargetRoomName, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetZoneAttributes action.
        """
        arguments = {
            "DesiredZoneName": DesiredZoneName,
            "DesiredIcon": DesiredIcon,
            "DesiredConfiguration": DesiredConfiguration,
            "DesiredTargetRoomName": DesiredTargetRoomName,
        }

        self.call_action("SetZoneAttributes", arguments=arguments, aspects=aspects)

        return




from typing import Dict, List, Optional

import os
import threading

import zeroconf

from akit.xlogging.foundations import AKitLoggerWrapper, AKitLogLevels

from akit.interop.dns.mdnsserviceinfo import MdnsServiceInfo
from akit.waiting import WaitContext

class MdnsServiceCatalog(zeroconf.ServiceListener):
    """
    """

    def __init__(self, logger: Optional[AKitLoggerWrapper]=None, log_level: AKitLogLevels=AKitLogLevels.INFO):
        self._logger = logger
        self._log_level = log_level
        
        self._lock = threading.Lock()
        self._service_catalog: Dict[str, Dict[str, zeroconf.ServiceInfo]] = {}
        return

    def update_service(self, zc: zeroconf.Zeroconf, svc_type: str, svc_name: str) -> None:

        svc_info = MdnsServiceInfo(zc.get_service_info(svc_type, svc_name))
        
        self._lock.acquire()
        try:
            svc_name_catalog = self._pull_svcname_catalog_for_svctype(svc_type)
            svc_name_catalog[svc_name] = svc_info
        finally:
            self._lock.release()

        self._log_service_activity(svc_type, svc_name, svc_info, "Service update type={} name={}.")

        return
    
    def remove_service(self, zc: zeroconf.Zeroconf, svc_type: str, svc_name: str) -> None:

        svc_info = None

        self._lock.acquire()
        try:
            svc_name_catalog = self._pull_svcname_catalog_for_svctype(svc_type)
            svc_info = svc_name_catalog[svc_name]
            del svc_name_catalog[svc_name]
        finally:
            self._lock.release()

        self._log_service_activity(svc_type, svc_name, svc_info, "Service remove type={} name={}.")

        return

    def add_service(self, zc: zeroconf.Zeroconf, svc_type: str, svc_name: str) -> None:
        
        svc_info = MdnsServiceInfo(zc.get_service_info(svc_type, svc_name))
    
        self._lock.acquire()
        try:
            svc_name_catalog = self._pull_svcname_catalog_for_svctype(svc_type)
            svc_name_catalog[svc_name] = svc_info
        finally:
            self._lock.release()

        self._log_service_activity(svc_type, svc_name, svc_info, "Service add type={} name={}.")

        return

    def list_service_names_for_type(self, svc_type: str) -> List[str]:

        svc_name_list = None
        
        self._lock.acquire()
        try:
            svc_name_catalog = self._pull_svcname_catalog_for_svctype(svc_type)
            svc_name_list = [ svc_name for svc_name in svc_name_catalog.keys()]
        finally:
            self._lock.release()
    
        svc_name_list.sort()

        return svc_name_list

    def lookup_service_catalog_for_type(self, svc_type: str) -> MdnsServiceInfo:
        
        svc_name_catalog = {}

        self._lock.acquire()
        try:
            # Create a copy of the service catalog to return so we don't create a issue
            # with multiple thread sharing a dict collection type
            svc_name_catalog.update(self._pull_svcname_catalog_for_svctype(svc_type))
        finally:
            self._lock.release()

        return svc_name_catalog

    def lookup_service_info(self, svc_type: str, svc_name: str) -> MdnsServiceInfo:
        
        service_info = None

        self._lock.acquire()
        try:
            svc_name_catalog = self._pull_svcname_catalog_for_svctype(svc_type)

            if svc_name in svc_name_catalog:
                service_info = svc_name_catalog[svc_name]
        finally:
            self._lock.release()

        return service_info

    def wait_for_services_in_service_catalog(self, wctx: WaitContext, zc: zeroconf.Zeroconf, svc_type: str, svc_query_names: List[str], exact_match=False):

        all_present = True

        if wctx.is_do_every_interval:
            # If this is a do every interval, send out another copy of our DNSQuestion
            out = zeroconf.DNSOutgoing(zeroconf.const._FLAGS_QR_QUERY, multicast=True)
            question = zeroconf.DNSQuestion(svc_type, zeroconf.const._TYPE_PTR, zeroconf.const._CLASS_IN)
            out.add_question(question)
            zc.send(out)
            
        service_catalog = self.lookup_service_catalog_for_type(svc_type)

        if exact_match:
            for exp_svc_name in svc_query_names:

                item_found = False
                for svc_name in service_catalog:
                    if svc_name == exp_svc_name:
                        item_found = True
                        break

                if not item_found:
                    all_present = False
                    break
        else:
            for exp_svc_name in svc_query_names:

                item_found = False
                for svc_name in service_catalog:
                    if svc_name.startswith(exp_svc_name):
                        item_found = True
                        break

                if not item_found:
                    all_present = False
                    break

        if not all_present and wctx.final_attempt:
            what_for = "Timeout waiting for sevice discovery for mdns services:"

            detail_lines = [
                "SERVICE NAMES:"
            ]
            for sqname in svc_query_names:
                detail_lines.append(f"    {sqname}")

            toerr = wctx.create_timeout(what_for, detail_lines)
            raise toerr

        return all_present

    def _log_service_activity(self, svc_type: str, svc_name: str, svc_info: MdnsServiceInfo, what_template: str):

        msg_lines = [
            what_template.format(svc_type, svc_name)
        ]

        if self._log_level <= AKitLogLevels.DEBUG:
            msg_lines.extend(svc_info.detail_lines())
            msg_lines.append("")

        msg = os.linesep.join(msg_lines)

        if self._logger is not None:
            self._logger.info(msg)
        else:
            print(msg)

        return

    def _pull_svcname_catalog_for_svctype(self, svc_type: str) -> Dict[str, MdnsServiceInfo]:

        svc_name_catalog = None
        if svc_type not in self._service_catalog:
            svc_name_catalog = {}
            self._service_catalog[svc_type] = svc_name_catalog
        else:
            svc_name_catalog = self._service_catalog[svc_type]
        
        return svc_name_catalog



"""
.. module:: xcollections
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module which contains classes that are enhancements of the collections classes

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"


from typing import Any

def insert_into_ordered_list_ascending(ordered_list: list, item: Any):
    """
        Takes a list of similar items that are already sorted in ascending order
        and inserts the 'item' parameter into the list in the correct order.

        :param ordered_list: The sorted list to insert into.
        :param item: The item to insert into the list.
    """
    ordered_list = []
    index = None
    for idx, nxt in enumerate(ordered_list):
        if nxt > item:
            index = idx

    if index is not None:
        ordered_list.insert(index, item)
    else:
        ordered_list.append(item)

    return


class CaseInsensitiveBytesDict(dict):
    """
        The :class:`CaseInsensitiveBytesDict` is a dictionary that can store and
        recover values associated with a text based key in a case insensative
        way.  It accomplishes this by converting all ascii characters in a key
        to lower case.
    """

    @staticmethod
    def key_to_lower(key: bytes):
        """
            Converts a key to a lower case string so case insensative comparisons can be made.

            :param key: A key value to check to see if an object is stored under the key.
        """
        if not isinstance(key, bytes):
            raise KeyError("Only bytes keys can be used with the CaseInsensitiveBytesDict type.")

        key = key.lower()
        return key

    def __init__(self, *args, **kwargs):
        super(CaseInsensitiveBytesDict, self).__init__(*args, **kwargs)
        self._convert_keys()
        return

    def __contains__(self, key: bytes) -> bool:
        """
            Overrides the :method:`__contains__` operator providing a key conversion in order
            to perform a case insensative dictionary key lookup.

            :param key: A key value to check to see if an object is stored under the key.

            :returns: A boolean indicating if the key provided is in this objects keys.
        """
        return super(CaseInsensitiveBytesDict, self).__contains__(CaseInsensitiveBytesDict.key_to_lower(key))

    def __delitem__(self, key: bytes):
        """
            Overrides the :method:`__delitem__` operator providing a key conversion in order
            to perform a case insensative dictionary item delete.

            :param key: A key value to delete the object stored under the key.

            :raises: KeyError
        """
        return super(CaseInsensitiveBytesDict, self).__delitem__(CaseInsensitiveBytesDict.key_to_lower(key))

    def __getitem__(self, key: bytes) -> Any:
        """
            Overrides the :method:`__getitem__` operator providing a key conversion in order
            to perform a case insensative dictionary item lookup.

            :param key: A key value to lookup and return the object stored under the key.

            :returns: The object associated with the key provided.

            :raises: KeyError
        """
        return super(CaseInsensitiveBytesDict, self).__getitem__(CaseInsensitiveBytesDict.key_to_lower(key))

    def __setitem__(self, key: bytes, value: Any):
        """
            Overrides the :method:`__setitem__` operator providing a key conversion in order
            to perform a case insensative dictionary item set.

            :param key: A key value to use to store the object stored under the key.
            :param value: The value to store with the key.
        """
        super(CaseInsensitiveBytesDict, self).__setitem__(CaseInsensitiveBytesDict.key_to_lower(key), value)

    def pop(self, key: bytes, *args, **kwargs) -> Any:
        """
            Overrides the :method:`pop` method providing a key conversion in order
            to perform a case insensative dictionary pop item.

            :param key: A key value to use to pop and return the object stored under the key.

            :returns: The object associated with the key provided.

            :raises: KeyError
        """
        return super(CaseInsensitiveBytesDict, self).pop(CaseInsensitiveBytesDict.key_to_lower(key), *args, **kwargs)

    def get(self, key: bytes, *args, **kwargs) -> Any:
        """
            Overrides the :method:`get` method providing a key conversion in order
            to perform a case insensative dictionary get item.

            :param key: A key value to use to get and return the object stored under the key.

            :returns: The object associated with the key provided.

            :raises: KeyError
        """
        return super(CaseInsensitiveBytesDict, self).get(CaseInsensitiveBytesDict.key_to_lower(key), *args, **kwargs)

    def setdefault(self, key: bytes, value: Any, *args, **kwargs):
        """
            Overrides the :method:`setdefault` method providing a key conversion in order
            to perform a case insensative dictionary setdefault for the given key.

            :param key: A key value to use to store the object stored under the key.
            :param value: The default value to store with the key.
        """
        return super(CaseInsensitiveBytesDict, self).setdefault(CaseInsensitiveBytesDict.key_to_lower(key), value, *args, **kwargs)

    def update(self, dobj: dict = {}, **kwargs): # pylint: disable=dangerous-default-value
        """
            Overrides the :method:`update` method providing a key conversion in order
            to perform a case insensative dictionary update.
        """
        super(CaseInsensitiveBytesDict, self).update(self.__class__(dobj))
        super(CaseInsensitiveBytesDict, self).update(self.__class__(**kwargs))
        return

    def _convert_keys(self):
        """
            Used by the :class:`CaseInsensitiveBytesDict` constructor to convert the keys of
            initialization dictionaries to lower case keys so we can perform case insensative
            key lookups.
        """
        for key in list(self.keys()):
            val = super(CaseInsensitiveBytesDict, self).pop(key)
            self.__setitem__(key, val)
        return



class CaseInsensitiveStringDict(dict):
    """
        The :class:`CaseInsensitiveStringDict` is a dictionary that can store and
        recover values associated with a text based key in a case insensative
        way.  It accomplishes this by converting all ascii characters in a key
        to lower case.
    """

    @staticmethod
    def key_to_lower(key: str):
        """
            Converts a key to a lower case string so case insensative comparisons can be made.

            :param key: A key value to check to see if an object is stored under the key.
        """
        if not isinstance(key, str):
            raise KeyError("Only string keys can be used with the CaseInsensitiveStringDict type.")

        key = key.lower()
        return key

    def __init__(self, *args, **kwargs):
        super(CaseInsensitiveStringDict, self).__init__(*args, **kwargs)
        self._convert_keys()
        return

    def __contains__(self, key: str) -> bool:
        """
            Overrides the :method:`__contains__` operator providing a key conversion in order
            to perform a case insensative dictionary key lookup.

            :param key: A key value to check to see if an object is stored under the key.

            :returns: A boolean indicating if the key provided is in this objects keys.
        """
        return super(CaseInsensitiveStringDict, self).__contains__(CaseInsensitiveStringDict.key_to_lower(key))

    def __delitem__(self, key: str):
        """
            Overrides the :method:`__delitem__` operator providing a key conversion in order
            to perform a case insensative dictionary item delete.

            :param key: A key value to delete the object stored under the key.

            :raises: KeyError
        """
        return super(CaseInsensitiveStringDict, self).__delitem__(CaseInsensitiveStringDict.key_to_lower(key))

    def __getitem__(self, key: str) -> Any:
        """
            Overrides the :method:`__getitem__` operator providing a key conversion in order
            to perform a case insensative dictionary item lookup.

            :param key: A key value to lookup and return the object stored under the key.

            :returns: The object associated with the key provided.

            :raises: KeyError
        """
        return super(CaseInsensitiveStringDict, self).__getitem__(CaseInsensitiveStringDict.key_to_lower(key))

    def __setitem__(self, key: str, value: Any):
        """
            Overrides the :method:`__setitem__` operator providing a key conversion in order
            to perform a case insensative dictionary item set.

            :param key: A key value to use to store the object stored under the key.
            :param value: The value to store with the key.
        """
        super(CaseInsensitiveStringDict, self).__setitem__(CaseInsensitiveStringDict.key_to_lower(key), value)

    def pop(self, key: str, *args, **kwargs) -> Any:
        """
            Overrides the :method:`pop` method providing a key conversion in order
            to perform a case insensative dictionary pop item.

            :param key: A key value to use to pop and return the object stored under the key.

            :returns: The object associated with the key provided.

            :raises: KeyError
        """
        return super(CaseInsensitiveStringDict, self).pop(CaseInsensitiveStringDict.key_to_lower(key), *args, **kwargs)

    def get(self, key: str, *args, **kwargs) -> Any:
        """
            Overrides the :method:`get` method providing a key conversion in order
            to perform a case insensative dictionary get item.

            :param key: A key value to use to get and return the object stored under the key.

            :returns: The object associated with the key provided.

            :raises: KeyError
        """
        return super(CaseInsensitiveStringDict, self).get(CaseInsensitiveStringDict.key_to_lower(key), *args, **kwargs)

    def setdefault(self, key: str, value: Any, *args, **kwargs):
        """
            Overrides the :method:`setdefault` method providing a key conversion in order
            to perform a case insensative dictionary setdefault for the given key.

            :param key: A key value to use to store the object stored under the key.
            :param value: The default value to store with the key.
        """
        return super(CaseInsensitiveStringDict, self).setdefault(CaseInsensitiveStringDict.key_to_lower(key), value, *args, **kwargs)

    def update(self, dobj: dict = {}, **kwargs): # pylint: disable=dangerous-default-value
        """
            Overrides the :method:`update` method providing a key conversion in order
            to perform a case insensative dictionary update.
        """
        super(CaseInsensitiveStringDict, self).update(self.__class__(dobj))
        super(CaseInsensitiveStringDict, self).update(self.__class__(**kwargs))
        return

    def _convert_keys(self):
        """
            Used by the :class:`CaseInsensitiveStringDict` constructor to convert the keys of
            initialization dictionaries to lower case keys so we can perform case insensative
            key lookups.
        """
        for key in list(self.keys()):
            val = super(CaseInsensitiveStringDict, self).pop(key)
            self.__setitem__(key, val)
        return



"""
.. module:: activate
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module that is utilized by test files to ensure the test environment is initialized in
               the correct order.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import os
import sys

from datetime import datetime
from akit.exceptions import AKitConfigurationError, AKitSemanticError, TRACEBACK_CONFIG, VALID_MEMBER_TRACE_POLICY

# Perform a sematic check to see who is importing the akit.activation.base module.  We
# need to make sure that the user is following the proper semantics and importing an activation
# profile and not directly importing this module.  This will enforce the setting of the
# activation profile in the global variables and enforce a proper environment activation
# sequence is followed.
importer_frame = sys._getframe()
while True:
    importer_frame = importer_frame.f_back
    if importer_frame.f_code.co_filename.find("importlib") < 0:
        break
if "__activation_profile__" not in importer_frame.f_locals:
    errmsg = "The 'akit.activation.base' should not be directly imported." \
             "  The environment activation should always happen by importing" \
             " an activation profile module."
    raise AKitSemanticError(errmsg)

# =======================================================================================
# The way we start up the test framework and the order which things come up in is a very
# important part of the automation process.  It effects whether or not logging is brought
# up consistently before all modules start using it.  It ensures that no matter how we
# enter into an automation process, whether via a test runner, terminal, or debugging a single
# file that we properly parse arguments and settings and launch the automation process
# consistently.
#
# Because of these necessities, we setup the activate module so it is the first thing
# scripts and tests files that consume the test framework will import to ensure the
# startup process is always consistent
#
# The framework has a special activation module :module:`akit.environment.console` that is
# used when bringing up the test framework in a console.  This special method redirects

# Activation Step - 1: Force the CONFIGURATION_MAP global variable into existance and
# pull in the OVERRIDE_CONFIGURATION dictionary.
from akit.environment.configuration import OVERRIDE_CONFIGURATION, CONFIGURATION_MAP

# Activation Step - 2: Force the global shared context to load, we want this to happen as early
# as possible because we don't want to every replace its reference or invalidate
# any references to it that someone might have acquired.
from akit.environment.context import Context # pylint: disable=wrong-import-position

ctx = Context()

# Activation Step - 3: Process the environment variable overrides for any of the AKIT configuration
# variables. This needs to happen before we load or create an initial user configuration
# because the variables may effect the values we write into the user configuration file.
from akit.environment.variables import (
    AKIT_VARIABLES,
    JOB_TYPES,
    
    LOG_LEVEL_NAMES,
    normalize_variable_whitespace
)

# Activation Step - 4: Load the user and runtime configuration and add it to the CONFIGURATION_MAP
# 'ChainMap' so the runtime settings can take precedence over the user default settings. 
from akit.environment.configuration import load_runtime_configuration, load_user_configuration

AKIT_DIR = AKIT_VARIABLES.AKIT_DIR

DEFAULT_PATH_EXPANSIONS = [
    os.path.expanduser,
    os.path.expandvars,
    os.path.abspath
]
def expand_path(path_in, expansions=DEFAULT_PATH_EXPANSIONS):

    path_out = path_in
    for expansion_func in expansions:
        path_out = expansion_func(path_out)

    return path_out


# The runtime configuration should be first so it
# has the highest priority.
runtime_config = load_runtime_configuration()
CONFIGURATION_MAP.maps.append(runtime_config)

# User config is last in the map list because runtime
# overrides user configuration, we do it this way because
# the runtime configuration might be generated by the
# CI system.  The user configuration is persisted and
# can have settings that stay with the machine.
user_config = load_user_configuration()
CONFIGURATION_MAP.maps.append(user_config)


# Activation Step - 5: Write the information into the context that is not persisted and is
# instance specific.  These variables are stored in the environment section
# of the configuration and overall global context.
env = ctx.lookup("/environment")

env["apod"] = AKIT_VARIABLES.AKIT_APOD_NAME

env["build"] = {
    "branch": AKIT_VARIABLES.AKIT_BUILD_BRANCH,
    "name": AKIT_VARIABLES.AKIT_BUILD_NAME,
    "flavor": AKIT_VARIABLES.AKIT_BUILD_FLAVOR,
    "url": AKIT_VARIABLES.AKIT_BUILD_URL,
}

env["job"] = {
    "initiator": AKIT_VARIABLES.AKIT_JOB_INITIATOR,
    "label": AKIT_VARIABLES.AKIT_JOB_LABEL,
    "name": AKIT_VARIABLES.AKIT_JOB_NAME,
    "owner": AKIT_VARIABLES.AKIT_JOB_OWNER,
    "type": AKIT_VARIABLES.AKIT_JOBTYPE
}

# Set the jobtype on the base environment for convenience
env["jobtype"] = AKIT_VARIABLES.AKIT_JOBTYPE

env["breakpoints"] = AKIT_VARIABLES.AKIT_BREAKPOINTS
env["debugger"] = AKIT_VARIABLES.AKIT_DEBUGGER
env["testroot"] = AKIT_VARIABLES.AKIT_TESTROOT
env["starttime"] = AKIT_VARIABLES.AKIT_STARTTIME
env["runid"] = AKIT_VARIABLES.AKIT_RUNID
env["pid"] = os.getpid()



# We set all the variables for config file options from the environment
# we just loaded, these might get overridden late but that is ok

configuration = ctx.lookup("/configuration", default={})
configuration["skip-devices"] = []

if "diagnostics" not in configuration:
    configuration["diagnostics"] = {}

if AKIT_VARIABLES.AKIT_TRACEBACK_POLICY_OVERRIDE is not None:
    if AKIT_VARIABLES.AKIT_TRACEBACK_POLICY_OVERRIDE not in VALID_MEMBER_TRACE_POLICY:
        configuration["diagnostics"]["traceback-policy-override"] = AKIT_VARIABLES.AKIT_TRACEBACK_POLICY_OVERRIDE
    else:
        errmsg = "Invalid traceback policy environment value. AKIT_TRACEBACK_POLICY_OVERRIDE={}".format(
            AKIT_VARIABLES.AKIT_TRACEBACK_POLICY_OVERRIDE
        )
        raise AKitConfigurationError(errmsg)

# If a traceback policy override was set, apply it in the TRACEBACK_CONFIG in the akit.exceptions module
if "traceback-policy-override" in configuration["diagnostics"]:
    TRACEBACK_CONFIG.TRACEBACK_POLICY_OVERRIDE = configuration["diagnostics"]["traceback-policy-override"]

config_files = configuration.lookup("/paths", default={})
config_files["credentials"] = AKIT_VARIABLES.AKIT_CONFIG_CREDENTIALS
config_files["landscape"] = AKIT_VARIABLES.AKIT_CONFIG_LANDSCAPE
config_files["runtime"] = AKIT_VARIABLES.AKIT_CONFIG_RUNTIME
config_files["topology"] = AKIT_VARIABLES.AKIT_CONFIG_TOPOLOGY
config_files["user"] = AKIT_VARIABLES.AKIT_CONFIG_USER

# Activation Step - 5: After 
if AKIT_VARIABLES.AKIT_LOG_LEVEL_CONSOLE is not None and AKIT_VARIABLES.AKIT_LOG_LEVEL_CONSOLE in LOG_LEVEL_NAMES:
    console_level = AKIT_VARIABLES.AKIT_LOG_LEVEL_CONSOLE
else:
    console_level = "INFO"
    AKIT_VARIABLES.AKIT_LOG_LEVEL_CONSOLE = console_level

if AKIT_VARIABLES.AKIT_LOG_LEVEL_FILE is not None and AKIT_VARIABLES.AKIT_LOG_LEVEL_FILE in LOG_LEVEL_NAMES:
    logfile_level = AKIT_VARIABLES.AKIT_LOG_LEVEL_FILE
else:
    logfile_level = "DEBUG"
    AKIT_VARIABLES.AKIT_LOG_LEVEL_FILE = logfile_level

if AKIT_VARIABLES.AKIT_SKIP_DEVICES is not None:
    devices_list = normalize_variable_whitespace(AKIT_VARIABLES.AKIT_SKIP_DEVICES).split(" ")
    configuration["skip-devices-override"] = devices_list

fill_dict = {
    "starttime": str(AKIT_VARIABLES.AKIT_STARTTIME).replace(" ", "T")
}

jobtype = ctx.lookup("/environment/job/type", default=JOB_TYPES.TESTRUN)

# We want to pull the console and testresults value from the configuration, because if its not there it
# will be set from the default_dir_template variable
default_dir_template = os.path.join(AKIT_VARIABLES.AKIT_HOME_DIRECTORY, "results", "console", "%(starttime)s")
outdir_template_consoleresults = configuration.lookup("/path-templates/consoleresults", default=default_dir_template)
dir_consoleresults = default_dir_template % fill_dict
configuration.insert("/paths/consoleresults", dir_consoleresults)

default_dir_template = os.path.join(AKIT_VARIABLES.AKIT_HOME_DIRECTORY, "results", "testresults", "%(starttime)s")
outdir_template_testresults = configuration.lookup("/path-templates/testresults", default=default_dir_template)
dir_testresults = default_dir_template % fill_dict
configuration.insert("/paths/testresults", dir_testresults)

outdir_full = None
# Figure out which output directory to set as the current process output directory.  The output directory
# determines where logging will go and is different depending on the activation mode of the test framework
if AKIT_VARIABLES.AKIT_OUTPUT_DIRECTORY is not None:
    outdir_full = expand_path(AKIT_VARIABLES.AKIT_OUTPUT_DIRECTORY % fill_dict)
    env["output_directory"] = outdir_full

    if jobtype == JOB_TYPES.TESTRUN:
        configuration.insert("/paths/testresults", outdir_full)
    elif jobtype == JOB_TYPES.CONSOLE:
        configuration.insert("/paths/testresults", outdir_full)
else:
    if jobtype == JOB_TYPES.CONSOLE:
        outdir_full = expand_path(outdir_template_consoleresults % fill_dict)
        env["output_directory"] = outdir_full
    else:
        outdir_full = expand_path(outdir_template_testresults % fill_dict)
        env["output_directory"] = outdir_full

results_configuration = {}
results_configuration["static-resource-dest-dir"] = expand_path(AKIT_VARIABLES.AKIT_RESULTS_STATIC_RESOURCE_DEST_DIR)
results_configuration["static-resource-src-dir"] = expand_path(AKIT_VARIABLES.AKIT_RESULTS_STATIC_RESOURCE_SRC_DIR)

if AKIT_VARIABLES.AKIT_RESULTS_HTML_TEMPLATE is not None:
    results_configuration["html-template"] = AKIT_VARIABLES.AKIT_RESULTS_HTML_TEMPLATE
else:
    results_configuration["html-template"] = expand_path(os.path.join(AKIT_DIR, "templates", "testsummary.html"))

configuration["results-configuration"] = results_configuration

if jobtype != "console":
    env["behaviors"] = {
        "log-landscape-declaration": True,
        "log-landscape-scan": True
    }

# Activation Step - 7: Import the logging module so we can be the trigger the logging configuration
# for standard out
import akit.xlogging.foundations # pylint: disable=unused-import,wrong-import-position



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Any, Optional, Tuple

import weakref

from datetime import datetime

from akit.interop.eventsinking.enumerations import EventedVariableState

class UpnpDefaultVar:
    """
        The UpnpDefaultVar object is utilized to handle the storage of the variables
        that the UpnpService has declared as fixed or not evented.
    """

    def __init__(self, key: str, name: str, service_ref: weakref.ref, value: Any = None, data_type: Optional[str] = None, default: Any = None,
                 allowed_list=None, timestamp: datetime = None, evented: bool = False):
        """
            Constructor for the :class:`UpnpEventVar` object.

            :param key: The key {service type}/{event name} for this event
            :param name: The name of the event this variable is storing information on.
            :param value: Optional initially reported value for the variable.  This is used when we have reports for
                          variables that we are not subscribed to.
            :param timestamp: The timestamp of the creation of this variable.  If a timestamp is passed then a value
                              needs to also be passed.
            :param evented: Indicates that this variable is evented to subscribers.
        """
        self._key = key
        self._name = name
        self._service_ref = service_ref
        self._value = value
        self._data_type = data_type
        self._default = default
        self._allowed_list = allowed_list
        self._timestamp = None
        
        if evented:
            errmsg = "UpnpDefaultVar constructor was called for a variabled that is evented."
            raise ValueError(errmsg)

        self._expires = None

        if self._value is not None and timestamp is None:
            self._timestamp  = datetime.now()

        if value is None and default is not None:
            self._value = default

        self._created = timestamp
        self._updated = timestamp
        self._changed = timestamp
        return

    @property
    def created(self) -> datetime:
        """
            When the event variabled value was set for the first time.
        """
        return self._created

    @property
    def changed(self) -> datetime:
        """
            A datetime object that indicates when the value was last changed in value.
        """
        return self._changed

    @property
    def expired(self) -> bool:
        """
            When the event variabled subscription has expired.
        """
        return False

    @property
    def key(self) -> str:
        """
            The key {service type}/{event name} for this event.
        """
        return self._key

    @property
    def state(self) -> EventedVariableState:
        """
            The state of this event variable, UnInitialized, Valid or Stale
        """
        return EventedVariableState.Default

    @property
    def updated(self) -> datetime:
        """
            A datetime object that indicates when the value was last updated.
        """
        return self._updated

    @property
    def name(self) -> str:
        """
            The name of the event this variable is storing information on.
        """
        return self._name

    @property
    def value(self) -> Any:
        """
            The last value reported for the event variable this instance is referencing.
        """
        return self._value

    def sync_read(self) -> Tuple[Any, datetime, datetime, EventedVariableState]:
        """
            Performs a threadsafe read of the value, updated, and state members of a
            :class:`UpnpEventVar` instance.
        """
        value, updated, changed, state = None, None, None, EventedVariableState.UnInitialized

        service = self._service_ref()
        for _ in service.yield_service_lock():
            updated = self._updated
            changed = self._changed

            state = EventedVariableState.Default

            value = self._value

        return value, updated, changed, state

    def sync_update(self, value: Any, expires: Optional[datetime] = None, service_locked: bool = False):
        """
            Peforms a threadsafe update of the value, updated and sid members of a
            :class:`UpnpEventVar` instance.
        """
        updated = datetime.now()

        if service_locked:
            orig_value = self._value
            self._value, self._updated = value, updated
            if orig_value != self._value:
                self._changed = updated
            if expires is not None:
                self._expires = expires
        else:
            service = self._service_ref()
            for _ in service.yield_state_lock():
                orig_value = self._value
                self._value, self._updated = value, updated
                if orig_value != self._value:
                    self._changed = updated
                if expires is not None:
                    self._expires = expires

        return

    def __str__(self) -> str:
        value, updated, changed, state = self.sync_read()
        rtnstr = "name={} value={} updated={} changed={} state={}".format(self._name, value, updated, changed, state.name)
        return rtnstr


from typing import Dict

from akit.exceptions import AKitSemanticError

from akit.coupling.integrationcoupling import IntegrationCoupling
from akit.testing.testplus.scopecoupling import ScopeCoupling

from akit.testing.testplus.registration.integrationsource import IntegrationSource
from akit.testing.testplus.registration.parameterorigin import ParameterOrigin
from akit.testing.testplus.registration.resourcescope import ResourceScope
from akit.testing.testplus.registration.resourcesource import ResourceSource
from akit.testing.testplus.registration.scopesource import ScopeSource
from akit.testing.testplus.registration.parameterorigin import ParameterOrigin

from akit.testing.testplus.testref import TestRef


class ResourceRegistry:
    """
        1. During test discovery process, the parameter decorators are executed and the resource
           identifier names and source functions are registered and available for consumption via
           a query interface or by the test sequencer when generating the test execution flow code.

           The source functions and identifiers being registered are stored in the scope tree in
           order to build data structure that allows for the resolution of scope for the parameters
           that and associated identifiers that will be consumed by tests during the execution of
           the test execution sequence flow. (This is an implicit step)

        2. After all the tests have been imported and test references have been created, we go through
           all of the test references and ensure we have a completely filled out scope tree for all of
           the tests.  As we go we mark scopes as relavant so we can later go back through the tree
           and prune scope nodes that are not relavant to executing a test.

        3. Walk through the entire scope tree and prune non-relevant nodes starting at the leaves and
           working backwards until a relavant node is reached or the root of the tree is reached.

        4. Iterate over the test references and for each test reference, walk through the now pruned
           scope tree utilizing a parameter visibility dictionary chain assigning usages to each
           ParameterOrigin object as a reference is found.  As we walk, we make sure that any parameters
           reference in the current scope are visible in the current parameter dictionary and if not
           store information about the missing parameter.

        5. Go through all of the test references and for each test reference, collect integration and
           scope couplings that are referenced so we can build a combined picture of everything that
           is being coupled into the test run.
    """
    _instance = None
    _initialized = False

    def __new__(cls):
        """
            Constructs new instances of the ResourceRegistry object.
        """
        if cls._instance is None:
            cls._instance = super(ResourceRegistry, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        """
            Initializes the SubscriptionRegistry object the first time this singleton is created.
        """
        thisType = type(self)

        if not thisType._initialized:
            thisType._initialized = True

            # The source tables are how integration, scope and resource sources are
            # registered and they help to lookup any factory functions that generate
            # a given type of resource.
            self._integration_source = {}
            self._resource_source = {}
            self._scope_source = {}

            # The unknown parameter table gets populated with the names of subscribe functions
            # that have unknown paramters and the list of parameters names the could not be
            # resolved to a parameter source function.
            self._unknown_parameters = {}

            # The identifier subscription table contains a list of identifiers that are used by tests and
            # the associated contexts in which they were used.
            self._identifier_subscription_table = {}

            #TODO: Drop this once we 

            # The scope tree is used to keep a list of scopes associated with a
            # test run and the variable names that originate from a given scope.
            self._scope_tree_root = ResourceScope()

            self._referenced_integrations = {}
            self._referenced_scopes = {}
        return

    @property
    def referenced_integrations(self):
        return self._referenced_integrations

    @property
    def referenced_scopes(self):
        return self._referenced_scopes

    @property
    def unknown_parameters(self):
        return self._unknown_parameters

    def rename_resource_origins_from_main(self, new_origin):
        """
            Walks the scope tree and corrects the resource origin name for resources.

            ..note: This is a special use method that is typically use to replace the
            name '__main__' in resources loaded from a test file that is being debugged
            from a script entry point.
        """
        self._scope_tree_root.rename_resource_origins_from_main(new_origin)
        return

    def lookup_resource_scope(self, scope_name):
        resource_scope = self._scope_tree_root.lookup_scope(scope_name)
        return resource_scope

    def lookup_resource_source(self, source_func):

        source = None

        if source_func in self._integration_source:
            source = self._integration_source[source_func]
        elif source_func in self._scope_source:
            source = self._scope_source[source_func]
        elif source_func in self._resource_source:
            source = self._resource_source[source_func]

        return source

    def register_integration_source(self, source: IntegrationSource):
        """
            This method is called by the 'integration' decorator in order to register a
            factory function that generates an 'IntegrationCoupling' object. 
        """
        source_func = source.source_function

        if source_func not in self._integration_source:
            self._integration_source[source_func] = source

        return

    def register_parameter_origin(self, identifier: str, origin: ParameterOrigin):
        """
            The :method:`register_parameter_origin` is used to register an alias as a wellknown parameter
            so that subscriber functions can consume the parameter as an input parameter.
        """

        originating_scope = origin.originating_scope

        if self._scope_tree_root.has_descendent_parameter(originating_scope, identifier):
            errmsg = "A wellknown variable identified as '{}' has already been assigned to scope '{}'.".format(identifier, originating_scope)
            raise AKitSemanticError(errmsg) from None

        # Add the parameter origin to the identifiers_for_scope table for this scope so we
        # can lookup identifiers by scope
        self._scope_tree_root.add_descendent_parameter_origination(originating_scope, origin)

        return

    def register_resource_source(self, source: ResourceSource):
        """
            This method is called by the 'resource' decorator in order to register a
            factory function that generate an arbitrary parameter resources.
        """
        source_func = source.source_function
        
        if source_func not in self._resource_source:
            self._resource_source[source_func] = source

        return

    def register_scope_source(self, source: ScopeSource):
        """
            This method is called by the 'scope' decorator in order to register a factory
            function that generates an 'ScopeCoupling' object. 
        """
        source_func = source.source_function

        if source_func not in self._scope_source:
            self._scope_source[source_func] = source

        return

    def finalize_startup(self, test_references: Dict[str, TestRef]):
        """
            ..note: During test discovery process, the parameter decorators are executed and the resource
                    identifier names and source functions are registered and available for consumption via
                    a query interface or by the test sequencer when generating the test execution flow code.

                    The source functions and identifiers being registered are stored in the scope tree in
                    order to build data structure that allows for the resolution of scope for the parameters
                    that and associated identifiers that will be consumed by tests during the execution of
                    the test execution sequence flow. (This is an implicit step)
        """
        # 1. After all the tests have been imported and test references have been created, we go through
        #    all of the test references and ensure we have a completely filled out scope tree for all of
        #    the tests.  As we go we mark scopes as relavant so we can later go back through the tree
        #    and prune scope nodes that are not relavant to executing a test.
        for _, test_ref in test_references.items():
            self._scope_tree_root.ensure_parameter_scopes_for_test(test_ref)

        # 2. Walk through the entire scope tree and prune non-relevant nodes starting at the leaves and
        #    working backwards until a relavant node is reached or the root of the tree is reached.
        self._scope_tree_root.prune_unreference_scopes()

        # 3. Iterate over the test references and for each test reference, walk through the now pruned
        #    scope tree utilizing a parameter visibility dictionary chain assigning usages to each
        #    ParameterOrigin object as a reference is found.  As we walk, we make sure that any parameters
        #    reference in the current scope are visible in the current parameter dictionary and if not
        #    store information about the missing parameter.
        for _, test_ref in test_references.items():
            missing_params = []

            self._scope_tree_root.resolve_parameter_originations_for_test(test_ref, missing_params)

            if len(missing_params) > 0:
                self._unknown_parameters[test_ref.test_name] = missing_params

        # 4. Go through all of the test references and for each test reference, collect integration and
        #    scope couplings that are referenced so we can build a combined picture of everything that
        #    is being coupled into the test run.
        for _, test_ref in test_references.items():
            self._scope_tree_root.collect_integrations_and_scopes_for_test(self._referenced_integrations, self._referenced_scopes)

        return

resource_registry = ResourceRegistry()



"""
.. module:: shellscript
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: A module that provides the EmbeddedPython task class which implements
               the execution of inline python based tasks in a workpacket.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import List, Optional

import os

from akit.xformatting import indent_lines

from akit.workflow.tasks.taskbase import TaskBase

class EmbeddedPython(TaskBase):
    """
        A task class that is used to setup the running of an embedded python script
        in the context of the automation worker.
    """

    def __init__(self, ordinal: str, label: str, task_info: dict, logger):
        super(EmbeddedPython, self).__init__(ordinal, label, task_info, logger)
        self._script = task_info["script"]
        return

    @property
    def script(self) -> List[str]:
        return self._script

    def execute(self, parameters: Optional[dict]=None, topology: Optional[dict]=None, **kwargs) -> int:

        self._logger.info("STEP: %s - %s" % (self._label, self._ordinal))

        locals().update({"parameters": parameters, "topology": topology})
        locals().update(kwargs)

        script_content = os.linesep.join(self._script)
        script_content = indent_lines(script_content, level=3)

        exit_status = 0

        try:
            # Execute the inline python script in the context of the current
            # globals and locals.
            exec(script_content, globals(), locals())
        except:
            exit_status = 1

        return exit_status



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'source/packages'}

packages = \
['akit',
 'akit.activation',
 'akit.cli',
 'akit.cli.cmdtree',
 'akit.cli.cmdtree.databases',
 'akit.cli.cmdtree.generators',
 'akit.cli.cmdtree.network',
 'akit.cli.cmdtree.servers',
 'akit.cli.cmdtree.testing',
 'akit.coupling',
 'akit.datum',
 'akit.environment',
 'akit.interfaces',
 'akit.interop',
 'akit.interop.agents',
 'akit.interop.clients',
 'akit.interop.cluster',
 'akit.interop.concurrency',
 'akit.interop.coordinators',
 'akit.interop.credentials',
 'akit.interop.dns',
 'akit.interop.eventsinking',
 'akit.interop.landscaping',
 'akit.interop.landscaping.layers',
 'akit.interop.serial',
 'akit.interop.upnp',
 'akit.interop.upnp.content',
 'akit.interop.upnp.devices',
 'akit.interop.upnp.extensions',
 'akit.interop.upnp.extensions.dynamic',
 'akit.interop.upnp.extensions.dynamic.embeddeddevices',
 'akit.interop.upnp.extensions.dynamic.rootdevices',
 'akit.interop.upnp.extensions.dynamic.services',
 'akit.interop.upnp.extensions.dynamic.services.MicrosoftCorporation',
 'akit.interop.upnp.extensions.dynamic.services.NETGEAR',
 'akit.interop.upnp.extensions.dynamic.services.NETGEARInc',
 'akit.interop.upnp.extensions.dynamic.services.SonosInc',
 'akit.interop.upnp.extensions.standard.embeddeddevices',
 'akit.interop.upnp.extensions.standard.rootdevices',
 'akit.interop.upnp.extensions.standard.rootdevices.sonos',
 'akit.interop.upnp.extensions.standard.services',
 'akit.interop.upnp.extensions.standard.services.UPnP',
 'akit.interop.upnp.generator',
 'akit.interop.upnp.generator.dynamic',
 'akit.interop.upnp.services',
 'akit.interop.upnp.xml',
 'akit.interop.wireless',
 'akit.monitoring',
 'akit.networking',
 'akit.readers',
 'akit.readers.linux',
 'akit.readers.linux.command',
 'akit.templates',
 'akit.testing',
 'akit.testing.testplus',
 'akit.testing.testplus.registration',
 'akit.wellknown',
 'akit.workflow',
 'akit.workflow.tasks',
 'akit.xcryptography',
 'akit.xlogging',
 'akit.xml',
 'akit.xmultiprocessing',
 'akit.xregex',
 'akit.xthreading']

package_data = \
{'': ['*'],
 'akit.interop': ['soap/*'],
 'akit.interop.upnp.generator': ['standard/embeddeddevices/UPnP/*',
                                 'standard/services/UPnP/*'],
 'akit.interop.upnp.generator.dynamic': ['embeddeddevices/NETGEAR/*',
                                         'embeddeddevices/SonosInc/*',
                                         'rootdevices/MicrosoftCorporation/*',
                                         'rootdevices/NETGEAR/*',
                                         'rootdevices/NETGEARInc/*',
                                         'rootdevices/SonosInc/*',
                                         'services/MicrosoftCorporation/*',
                                         'services/NETGEAR/*',
                                         'services/NETGEARInc/*',
                                         'services/SonosInc/*'],
 'akit.monitoring': ['scripts/platforms/linux/*', 'templates/*'],
 'akit.templates': ['static/v0/*', 'tabs/*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'coverage>=7.0.4,<8.0.0',
 'debugpy>=1.6.5,<2.0.0',
 'dlipower>=1.0.176,<2.0.0',
 'flask-migrate>=4.0.1,<5.0.0',
 'flask-restx>=1.0.3,<2.0.0',
 'flask>=2.2.2,<3.0.0',
 'gunicorn>=20.1.0,<21.0.0',
 'ipython>=8.8.0,<9.0.0',
 'netifaces>=0.11.0,<0.12.0',
 'nmcli>=1.1.2,<2.0.0',
 'paramiko>=2.12.0,<3.0.0',
 'pycrunch-trace>=0.2,<0.3',
 'pylint>=2.15.9,<3.0.0',
 'pyreadline3>=3.4.1,<4.0.0',
 'pyserial>=3.5,<4.0',
 'pyyaml>=6.0,<7.0',
 'requests>=2.28.1,<3.0.0',
 'rich>=13.0.1,<14.0.0',
 'rope>=1.6.0,<2.0.0',
 'ssdp>=1.1.1,<2.0.0',
 'werkzeug>=2.2.2,<3.0.0',
 'zeroconf>=0.47.1,<0.48.0']

setup_kwargs = {
    'name': 'automationkit',
    'version': '0.3.61',
    'description': 'Automation Kit Test Framework',
    'long_description': '# Automation Kit Repository\n\nWelcome to the Automation Kit repository!\n\n\n*Note:* This package is progressing quickly but is not yet ready for full production environments.\n\n# Description\nThe **Automation Kit** is a python toolkit for building distributed automated task and testing systems.  It goes beyond the capabilities of other task and test automation frameworks by providing a hybrid approach that allows for dependency injection, similar to pytest, combined with and object oriented framework that can be utilized to easily customize the orchestration of tasks and tests across a collection of integrated resources.  Because the **Automation Kit** is built from the start with distributed automation in mind, it can help you quickly get distributed automated systems up and running that are based on a tested and proven set of object patterns and APIs.\n\n# Design Philosophies\nThis section covers some of the important design features of the **Automation Kit** framework.\n\n## Large Scale\nThe **Automation Kit** is designed for enterprise level distributed automation *"at scale"*.  The term *"at scale"* refers not only to a larger collection of enterprise resources but also refers to the fact that the **Automation Kit** helps to setup patterns that will support working in a very large code base.  The size of a code base that you might see associated with large enterprise level projects.  The **Automation Kit** supports working in large code bases by helping to establish good code organizational patterns and abstractions that support characteristics that:\n\n* make it easier to learn and work in the code base\n* make the code base easier to maintain\n* make it easier to share and reuse code\n\n## Faster Classification of Issues\nOne of the key philosophies behind the **Automation Kit** design is one of being able quickly and efficiently identify the nature of issues that come up during automation runs.  The **Automation Kit** initially classifies errors into one of four categories:\n\n* **Configuration** - We identify configuration issues quickly and classify them so as to ensure that resources are not waisted troubleshooting configuration related issues and that eronious test results or noise is not generated due to configuration related issues. We also want automation runs to fail out quickly if the automation configuration is not setup correctly, before waisting additional testing resources.\n* **Environment** - The **Automation Kit** performs an initial diagnostic scan of the automation landscape and all the resources declared to be necessary to run a series of tasks or tests in order to provide indications of environmental failures as early as possible.  This is important to ensure that we do not generate noise in automation results that are not related to the automation tasks or tests that might fail.  Just as with configuration, we want automation runs to fail out quickly if the automation environment is not setup correctly.\n* **Error** - The **Automation Kit** classifies un-expected errors, or errors that are not founded in an expectation of a result, as an **Error** condition.  This helps to ensure these errors are given an appropriate initial direction or indication that the issues is a problem in the automation code and not an issue in the code that is the target of the automation run.\n* **Failure** - The **AutomationKit** classifies failures that are associated with an expectation of behavior from a target under test as failures.  This allows for the proper initial classification of an issue as being a problem that is likely a failure in the code being targeted by the automation and the behavior or result it should have exhibited.\n\nHaving the initial classification of issues fall into one of these four categories helps to ensure that issues are easier to triage and assign to the appropriate party for investigation and resolution. It also helps to establish categories that can be used for data collection in order to better analyse the performance of test infrastructure, test code and product code.\n\n## Integration and Distributed Automation Support\nThe **Automation Kit** comes with enterprise level integration and distributed automation capabilities.  The framework utilizes a customize-able set of classes that guides enterprise users through a process of creating a very robust integration object model based on the roles that enterprise resources play in an automation landscape.\n\nThe declaration of a custom automation landscape is as simple as setting an environment variable or passing a command line flag declaring the python module that contains a custom landscape derived class.  The *Landscape* and *LandscapeDescription* derived classes work together to provide the **Automation Kit** with a description of the customized roles and integration mixin(s) that provide the connection between the tasks and test automation code.\n\n## Task and Test Integration Declaration and Assurance\nThe **Automation Kit** utilizes its object model to allow tasks and tests to provide information about their associated integration points and scopes of execution to the automation framework.  This integration declaration mechanism allows the automation framework to provide an early scan of the integration pathways and provide levels of assurance as to the stability of the automation landscape early in the automation process.  This is vitally important as it eliminates the waist and noise that are often associated with automation runs that are performed against an automation Landscape that has broken, mis-configured or missing resources.\n\n## Automation Job, Scope and Flow Control \nThe **Automation Kit** allows enterprise users to organize and customize the ordering of automation scope engagements and the flow of an automation job.  This provides the automation engineer the ability to control the engagement of automation scopes of execution and allows for optimal use of time and overlapping of scopes of execution in a test run.\n\n# Table of Contents\n1. [Automation Software Stack](http://automationmojo.com/automationkit/docs/usermanual/10-automation-software-stack.html)\n2. [Getting Started](http://automationmojo.com/automationkit/docs/usermanual/20-getting-started.html)\n3. [Automation Configuration](http://automationmojo.com/automationkit/docs/usermanual/30-automation-configuration.html)\n    1. [Landscape File](http://automationmojo.com/automationkit/docs/usermanual/31-landscape-file.html)\n    2. [Topology File](http://automationmojo.com/automationkit/docs/usermanual/32-topology-file.html)\n    2. [Runtime File](http://automationmojo.com/automationkit/docs/usermanual/33-runtime-file.html)\n    3. [Credentials File](http://automationmojo.com/automationkit/docs/usermanual/34-credentials-file.html)\n4. [TestRun Sequencing](http://automationmojo.com/automationkit/docs/usermanual/40-testrun-sequencing.html)\n    1. [Integration Couplings](http://automationmojo.com/automationkit/docs/usermanual/41-integration-couplings.html)\n    2. [Scope Couplings](http://automationmojo.com/automationkit/docs/usermanual/42-scope-couplings.html)\n5. [Functional Description](http://automationmojo.com/automationkit/docs/usermanual/50-functional-description.html)\n    1. [Activation and Startup](http://automationmojo.com/automationkit/docs/usermanual/51-activation-and-startup.html)\n    2. [Inter-Operability](http://automationmojo.com/automationkit/docs/usermanual/52-inter-operability.html)\n    3. [SSH Coordinator and Agent](http://automationmojo.com/automationkit/docs/usermanual/53-ssh-coordinator-and-agent.html)\n    4. [UPNP Coordinator and Agent](http://automationmojo.com/automationkit/docs/usermanual/54-upnp-coordinator-and-agent.html)\n6. [Workflow Orchestration](http://automationmojo.com/automationkit/docs/usermanual/60-workflow-orchestration.html)\n7. [Enterprise Extensibility](http://automationmojo.com/automationkit/docs/usermanual/70-enterprise-extensibility.html)\n8. [Command Line](http://automationmojo.com/automationkit/docs/usermanual/80-command-line.html)\n9. [Code Organization and Conventions](http://automationmojo.com/automationkit/docs/usermanual/90-code-organization-and-conventions.html)\n10. [Coding Standards](http://automationmojo.com/automationkit/docs/usermanual/100-coding-standards.html)\n',
    'author': 'Myron Walker',
    'author_email': 'myron.walker@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
